//-----------------------------------------------------------------------
// 
//  Copyright (C) Microsoft Corporation.  All rights reserved.
// 
//This source code is intended only as a supplement to Microsoft
//Development Tools and/or on-line documentation.  See these other
//materials for detailed information regarding Microsoft code samples.
// 
//THIS CODE AND INFORMATION ARE PROVIDED AS IS WITHOUT WARRANTY OF ANY
//KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
//IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
//PARTICULAR PURPOSE.
//-----------------------------------------------------------------------



using System;
using System.IO;
using System.Reflection;
using System.Collections;
using System.Text;

class TXRef
{
	static void Main(string[] args)
	{

		TypeWriter writer = new TypeWriter(Console.Out);
		foreach (string s in args)
		{
			Assembly a = null;
			try
			{
				a = Assembly.LoadFrom(s);
				if (a == null)
					a = Assembly.Load(s);
			}
			catch (Exception e) { Console.WriteLine(e); }
			if (a == null)
			{
				Console.WriteLine("TxRef: could not load assembly '{0}'.", s);
				return;
			}
			ArrayList l = new ArrayList();
			Type[] types;
			try
			{
				types = a.GetTypes();
			}
			catch (ReflectionTypeLoadException re)
			{
				foreach (Exception e in re.LoaderExceptions)
				{
					Console.WriteLine(e);
				}
				types = RemoveNulls(re.Types);

			}

			Array.Sort(types, new TypeComparer());

			foreach (Type t in types)
			{
				if (t.IsPublic)
				{
					if (!l.Contains(t.Namespace))
						l.Add(t.Namespace);
				}
			}

			l.Sort(null);

			foreach (string namespaceName in l)
			{
				writer.WriteNamespace(namespaceName, types);
			}
		}
	}

	static Type[] RemoveNulls(Type[] types)
	{
		ArrayList l = new ArrayList();
		foreach (Type t in types)
		{
			if (t != null) l.Add(t);
		}
		return (Type[])l.ToArray(typeof(Type));
	}
}

public class TypeComparer : IComparer
{
	internal enum TypeGroup {
		MainGroup = 0,
		Enums = 1,
		Collections = 2,
		DelegatesAndEventArgs = 3,
		Attributes = 4,
		Exceptions = 5
	}
	
	internal static TypeGroup GetGroupForType(Type type) {
		if (type.IsEnum) return TypeGroup.Enums;
		if ((typeof(IEnumerable).IsAssignableFrom(type))) return TypeGroup.Collections;
		if (type.IsSubclassOf(typeof(Delegate)) || type.IsSubclassOf(typeof(EventArgs))) return TypeGroup.DelegatesAndEventArgs;
		if (type.IsSubclassOf(typeof(Attribute))) return TypeGroup.Attributes;
		if (type.IsSubclassOf(typeof(Exception))) return TypeGroup.Exceptions;
		return TypeGroup.MainGroup;
	}

	public int Compare(object obj1, object obj2)
	{
		Type t1 = obj1 as Type;
		Type t2 = obj2 as Type;

		// these are used to group by (for example all enums together)
		TypeGroup t1Group = TypeComparer.GetGroupForType(t1);
		TypeGroup t2Group = TypeComparer.GetGroupForType(t2);

		if (t2Group != t1Group) return ((int)t1Group) < ((int)t2Group) ? -1 : 1;
		return t1.Name.CompareTo(t2.Name);
	}
}